#!/usr/bin/php
<?

  // default configuration settings
  $config['data_dir']        = "data";          // input data directory
  $config['work_dir']        = "work";          // temporary working directory
  $config['list_file_name']  = "plusminus.txt"; // plus-minus list file name
  $config['orgs_file_name']  = "organisms.txt"; // temp. organisms list file name
  $config['out_file_name']   = "results.txt";   // output file name
  $config['log_file_name']   = "protfile.log";  // log file name
  $config['scoremax_call']   = "scoremax";      // command to call 'scoremax' program
  $config['scoremax_param']  = "10";            // parameters to pass to 'scoremax'
  $config['max_results']     = 0;               // maximum number of shown results
  $config['resume_work']     = true;            // try to continue previous job
  $config['org_code_prefix'] = "ORG";           // prefix for organism codename
  $config['win_style_lf']    = false;           // lines end with \r\n

  date_default_timezone_set('Europe/Moscow');
  error_reporting( E_ALL & ~E_DEPRECATED );

  // saying hello
  echo "Protfile: Program for Specific Proteins Search\n".
       "(C) 2006, Laboratory of Mathematical Methods and Models in Bioinformatics\n".
       "Institute for Information Transmission Problems, Russian Academy of Sciences\n\n";

  // checking command line arguments if present
  if ( $argc > 1 )
  {
    // checking for help request or wrong number of parameters
    if ( $argc > 2 or ereg( "^-h|--help|/\?$", $argv[1] ) )
      die( "Usage:\n  protfile.php [<config_file_name>]\n" );

    // assuming the only parameter to be the config. file name
    $config_file_name = $argv[1]; // config. file name
    load_config($config_file_name); // loading config. from file
    add2log( "Loaded parameters from configuration file '$config_file_name'." );
  }

  // sending parameters list to log file
  add2log( "Current configuration parameters:\n" . config_list() );

  // checking for data directory presence
  if ( !file_exists($config['data_dir']) )
  {
    add2log( "No data directory exists." );
    die( "Can't find data directory '$config[data_dir]'." );
  }

  // checking for plus-minus list presence
  // if absent, creating template with all genomes available in data directory
  if ( !file_exists($config['list_file_name']) )
  {
    add2log( "No plus-minus list is found. Creating template." );
    $org_list = scan_directory_for_genomes($config['data_dir']);
    if ( empty($org_list) )
    {
      add2log( "No protein files were found in the data directory." );
      die( "No data found in the data directory '$config[data_dir]'." );
    }
    file_put_contents( $config['list_file_name'],
                       "[ ] " . implode( "\n[ ] ", $org_list ) . "\n"  );
    die( "Plus-minus list template created ('$config[list_file_name]'), ".
         "please fill it in." );
  }

  // reading plus-minus list
  add2log( "Reading plus-minus list from file '$config[list_file_name]'." );
  $organisms  = array(); // list of organism names
  $plus_list  = array(); // plus-lists
  $minus_list = array(); // minus-lists
  $first_list = array(); // 'first' list
  read_plusminus_list( $config['list_file_name'],
                       $organisms, $first_list, $plus_list, $minus_list );
  $org_codes = organisms_codenames(); // internal organisms codenames
  add2log( "Read " . count($organisms) . " organism names, " .
           count($plus_list) . " of them are in the plus-list, " .
           count($minus_list) . " - in the minus-list and " .
           count($first_list) . " will be treated as 'first'." );
  add2log( "Organism names with internal indexes:\n" . organisms_list() );
  add2log( "Plus-list:\n" . organisms_list($plus_list) );
  add2log( "Minus-list:\n" . organisms_list($minus_list) );
  add2log( "First-list:\n" . organisms_list($first_list) );

  // checking that all 3 lists are not empty
  if ( empty($plus_list) )
  {
    add2log( "Plus-list is empty." );
    die( "Plus-list must contain at least 1 organism. Stop." );
  }
  if ( empty($minus_list) )
  {
    add2log( "Minus-list is empty." );
    die( "Minus-list must contain at least 1 organism. Stop." );
  }
  if ( empty($first_list) )
  {
    add2log( "'First'-list is empty." );
    die( "At least 1 organism must be marked as 'first'. Stop." );
  }

  // checking if data from previous run(s) is available
  if ( $config['resume_work'] )
  {
    add2log( "Looking for data from previous run." );
    if ( !file_exists($config['work_dir']) ) // working directory does not exist
    {
      add2log( "The working directory '$config[work_dir]' does not exist." );
      $config['resume_work'] = false; // no working dir => no work to resume
    }
    elseif ( !glob("$config[work_dir]/*") ) // working directory is empty
    {
      add2log( "The working directory '$config[work_dir]' is empty." );
      $config['resume_work'] = false; // working dir is empty => no work to resume
    }
    elseif ( !file_exists($config['orgs_file_name']) ) // organisms list is absent
    {
      add2log( "No organisms list from previous run exists." );
      $config['resume_work'] = false; // no organisms list => no work to resume
    }
    elseif ( str_replace( "\r\n", "\n", file_get_contents($config['orgs_file_name']) ) != organisms_list() )
      // organisms list from previous run doesn't match current one
    {
      add2log( "Organisms list from previous run doesn't match current one. ".
               "Discarding it." );
      $config['resume_work'] = false;
    }
  }

  // if resume is unavailable or not needed, cleaning up the workspace
  // and preparing organisms list
  if ( $config['resume_work'] )
    add2log( "Using data from previous run." );
  else
  {
    if ( !file_exists($config['work_dir']) ) // working directory does not exist
    {
      add2log( "Creating working directory '$config[work_dir]'." );
      mkdir($config['work_dir']); // creating working directory
    }
    elseif ( $old_files = glob("$config[work_dir]/*") )
    {
      add2log( "Clearing data from previous run(s)." );
      foreach ( $old_files as $file )
        unlink($file);
    }

    if ( file_exists($config['orgs_file_name']) )
    {
      add2log( "Removing old organisms list file '$config[orgs_file_name]'." );
      unlink($config['orgs_file_name']);
    }

    add2log( "Creating organisms list '$config[orgs_file_name]'." );
    write_organisms_list($config['orgs_file_name']);
  }

  // checking existance of all needed genomes
  add2log( "Checking existance of all needed genomes." );
  foreach ( $organisms as $index => $organism )
    if ( in_array( $index, $plus_list ) or in_array( $index, $minus_list ) )
      if ( ( !file_exists("$config[work_dir]/$org_codes[$index].hdr") or
             !file_exists("$config[work_dir]/$org_codes[$index].seq") ) and
           !file_exists( "$config[data_dir]/$organism" ) )
      {
        add2log( "'$config[data_dir]/$organism' does not exist." );
        die( "Can't find genome directory for $organism." );
      }
  add2log( "All needed genome directories are present." );

  // preparing genomes
  echo "Preparing genomes ... ";
  add2log( "Preparing genomes." );
  foreach ( $organisms as $index => $organism )
    if ( in_array( $index, $plus_list ) or in_array( $index, $minus_list ) )
      if ( !file_exists("$config[work_dir]/$org_codes[$index].hdr") or
           !file_exists("$config[work_dir]/$org_codes[$index].seq") )
      {
        $faa_file_names = glob("$config[data_dir]/$organism/*.faa");
        if ( !$faa_file_names )
        {
          add2log( "There's no .faa files in '$config[data_dir]/$organism' directory." );
          die( "Can't find any .faa files for $organism." );
        }
        prepare_genome( $faa_file_names,
                        "$config[work_dir]/$org_codes[$index].hdr",
                        "$config[work_dir]/$org_codes[$index].seq" );
      }
  add2log( "Genomes prepared." );
  echo "done.\n";

  // openig output file
  if ( ! ($out_file = fopen( $config['out_file_name'], "wb" ) ) )
    die( "Can't open output file '$config[out_file_name]'." );

  $profiles   = array();
  $candidates = array();
  $pares_count = count($first_list)*(count($plus_list)+count($minus_list)-1);
  $current_pare = 0;
  foreach ( $first_list as $the_first )
  {
    echo "'First' genome - $org_codes[$the_first]\n";
    add2log( "'First' genome - $organisms[$the_first]." );
    my_fputs( $out_file, "'First' genome - $organisms[$the_first].\n" );

    $plus_list_ = $plus_list;
    unset( $plus_list_[array_search( $the_first, $plus_list_ )] );

    add2log( "Comparing genomes." );
    foreach ( $organisms as $the_second => $_ )
      if ( in_array( $the_second, $plus_list_ ) or
           in_array( $the_second, $minus_list ) )
      {
        $current_pare ++;
        $first_code  = $org_codes[$the_first];
        $second_code = $org_codes[$the_second];
        $score_file_name = "$config[work_dir]/$first_code-$second_code.scr";
        if ( file_exists( $score_file_name ) and
             read_scores( $score_file_name, $the_first, $the_second ) )
            continue;

        echo "Calculating SCOREs for $first_code vs. $second_code ".
             "($current_pare of $pares_count) ... ";
        add2log( "Calculating SCOREs for $organisms[$the_first] ($first_code) vs. ".
                                        "$organisms[$the_second] ($second_code)." );
        exec( "$config[scoremax_call] ".
              "$config[work_dir]/$first_code.seq ".
              "$config[work_dir]/$second_code.seq ".
              "$config[scoremax_param] ".
              "> $score_file_name" );

        if ( !file_exists( $score_file_name ) )
        {
          add2log( "'scoremax' didn't produce file '$score_file_name'." );
          die( "Can't load 'scoremax' results from '$score_file_name'." );
        }
    
        if ( !read_scores( $score_file_name, $the_first, $the_second ) )
        {
          add2log( "Improper 'scoremax' results in '$score_file_name'." );
          die( "Can't load 'scoremax' results from '$score_file_name'." );
        }
    
        echo "done.\n";
      }
    add2log( "Genomes comparison is complete." );

    echo "Analysing profiles ... ";
    add2log( "Calculating profiles quality." );
    $candidates[$the_first] = array();
    foreach ( $profiles[$the_first] as $prot_index => $profile )
      if ( minimax( $profile, $plus_list_ ) >
           maximax( $profile, $minus_list ) )
        $candidates[$the_first][] = profile_quality( $profile, $plus_list_, $minus_list );
    arsort($candidates[$the_first]);
    echo "done.\n";

    if ( empty($candidates[$the_first]) )
    {
      echo "No candidate proteins were found for $org_codes[$the_first].\n";
      add2log( "No condidates were found." );
      continue;
    }

    echo "Loading protein information headers for $org_codes[$the_first] ... ";
    add2log( "Loading protein information headers from file ".
             "'$config[work_dir]/$org_codes[$the_first].hdr'." );
    $headers = load_headers( "$config[work_dir]/$org_codes[$the_first].hdr" );
    echo "done.\n";

    echo "Producing results ... ";
    add2log( "Sending results to output file." );
    $num_results = 0;
    foreach ( $candidates[$the_first] as $prot_index => $quality )
    {
      my_fputs( $out_file, round( $quality, 4 ) . "\t$headers[$prot_index]\n" );
      if ( $config['max_results'] and ++$num_results > $config['max_results'] )
        break;
    }
    my_fputs( $out_file, "\n" );
    echo "done.\n";

  }

  fclose($out_file); // closing output file

  add2log( "All done." );
  echo "\nThank you for using the program.\n";



  /****** Functions ******/

  // prints a string to log file
  function add2log( $string )
  {
    global $config;

    static $log_file=null;

    if ( $log_file === null ) // log file isn't opened yet
      if ( ! ($log_file = fopen( $config['log_file_name'], "wb" ) ) )
        die( "Can't open log file '$config[log_file_name]'." );
      else
        add2log(  "`Protfile' program log file opened on " . date('Y-m-d') . "." );

    my_fputs( $log_file, date('[H:i:s]') . " $string\n" );

  }

  // replacement for fputs that converts line feeds to windows format if necessary
  function my_fputs( $file, $string )
  {
    global $config;
    if ( $config['win_style_lf'] )
    {
      $string = str_replace( "\n",     "\r\n", $string );
      $string = str_replace( "\r\r\n", "\r\n", $string );
    }
    return fputs( $file, $string );
  }

  // loads configuration parameters from file with given name
  function load_config( $filename )
  {

    global $config;

    if ( ! ($f = fopen( $filename, "rb" ) ) )
      die( "Can't open configuration file '$filename' for reading." );

    for ( $ln=1; ($line=fgets($f)) !== false; $ln++ )
      if ( strlen($line=trim($line)) )
        if ( !ereg( "^(#|;|//).*$", $line, $a ) ) // comment line
          if ( ereg( "^([A-Za-z0-9_]+) *= *\"? *(.*[^ ]) *\"?;?$", $line, $a ) )
          {
            $key   = $a[1];
            $value = $a[2];
            if ( array_key_exists( $key, $config ) )
            {
              if ( is_bool($config[$key]) and ( $value == "false" or $value == "no" ) )
                $value = false;
              settype( $value, gettype($config[$key]) );
              $config[$key] = $value;
            }
            else
              echo "Warning: unknown parameter $key on line $ln of configuration file.\n";
          }
          else
            echo "Warning: line $ln of configuration file was ignored.\n";

    fclose($f);
    
  }

  // generates a list of global configuration parameters, one per line
  function config_list()
  {
    global $config;

    $config_list = "";
    foreach ( $config as $key => $val )
    {
      if ( is_string( $val ) )
        $val = "'$val'";
      if ( $val === true )
        $val = 'true';
      if ( $val === false )
        $val = 'false';
      $config_list .= "$key = $val\n";
    }
    return $config_list;
  }

  // scans the given directory for directories containing faa files
  function scan_directory_for_genomes( $dir_name )
  {

    $org_list = array();
    $org_dirs = glob("$dir_name/*");
    foreach ( $org_dirs as $org_dir )
      if ( glob("$org_dir/*.faa") !== false )
        $org_list[] = substr( $org_dir, strlen($dir_name)+1 );
    return $org_list;

  }

  // generates a list of organisms with internal indexes and marks
  function organisms_list( $list=null )
  {
    global $organisms, $org_codes;

    $organisms_list = "";
    foreach ( $organisms as $index => $organism )
      if ( $list === null or in_array( $index, $list ) )
        $organisms_list .= "$org_codes[$index]: $organism\n";

    return $organisms_list;
  }

  // reads plus-minus list from file with given name
  function read_plusminus_list( $filename, &$organisms,
                                &$first_list, &$plus_list, &$minus_list )
  {

    $organisms  = array();
    $first_list = array();
    $plus_list  = array();
    $minus_list = array();

    if ( !($f=fopen( $filename, "rb" )) )
      die("Can't open plus-minus list file '$filename' for reading.");

    for ( $ln=1; ($line=fgets($f)) !== false; $ln++ )
      if ( strlen($line=trim($line)) )
        if ( ereg( "^ *\[([ *+-])\] *(.*) *$", $line, $a ) )
        {
          $mark = $a[1];
          $name = $a[2];
          $index = count($organisms);
          $organisms[$index] = $name;
          switch ( $mark )
          {
          case '*': $first_list[] = $index; // fall-through
          case '+': $plus_list[]  = $index; break;
          case '-': $minus_list[] = $index; break;
          }
        }

    fclose($f);

  }

  // makes organisms codenames from their serial numbers ^^
  function &organisms_codenames()
  {
    global $config, $organisms;
    $n_dig = strlen( count($organisms) );
    $org_codes = array();
    foreach ( $organisms as $index => $organism )
      $org_codes[$index] = sprintf( "$config[org_code_prefix]%0{$n_dig}d", $index+1 );
    return $org_codes;
  }

  // writes organisms list to file with specified name
  function write_organisms_list( $filename )
  {
    if ( !($f=fopen( $filename, "wb" )) )
      die("Can't open organisms list file '$filename' for writing.");
    my_fputs( $f, organisms_list() );
    fclose($f);
  }

  // gathers 1 or more .faa files and split them into headers file and sequences file
  function prepare_genome( $faa_file_names, $hdr_file_name, $seq_file_name )
  {

    $headers   = array();
    $sequences = array();

    foreach ( $faa_file_names as $faa_file_name )
    {
      if ( !($f=fopen( $faa_file_name, "rb" )) )
        die("Can't open .faa file '$faa_file_name' for reading.");

      for ( $ln=1, $pid=null, $sequence=''; ($line=fgets($f)) !== false; $ln++ )
        if ( strlen($line=trim($line)) )
          if ( $line[0] === '>' )
          {
            if ( $sequence )
              $sequences[] = $sequence;
            $sequence = '';

            if ( count($sequences) != count($headers) )
              die( "Error in FASTA file '$faa_file_name' on line $ln: ".
                   "headers count doesn't match sequences count." );

            $headers[] = substr( $line, 1 );
          }
          else
            $sequence .= $line;
      if ( $sequence )
        $sequences[] = $sequence;

      if ( count($sequences) != count($headers) )
        die( "Error in FASTA file '$faa_file_name' on line $ln: ".
             "headers count doesn't match sequences count." );

      if ( count($sequences) == 0 )
        die( "Error in FASTA file '$faa_file_name': no sequences." );

      fclose($f);
    }

    if ( !($hf=fopen( $hdr_file_name, "wb" )) )
      die("Can't open headers file '$hdr_file_name' for writing.");
    foreach ( $headers as $header )
      my_fputs( $hf, "$header\n" );
    fclose ($hf);

    if ( !($sf=fopen( $seq_file_name, "wb" )) )
      die("Can't open sequences file '$seq_file_name' for writing.");
    foreach ( $sequences as $sequence )
      my_fputs( $sf, "$sequence\n" );
    fclose ($sf);

  }

  // reads SCORE maximums from given file to global array $profiles
  function read_scores( $filename, $the_first, $the_second )
  {
    global $profiles;

    if ( !($f=fopen( $filename, "rb" )) )
      die("Can't open scores file '$filename' for reading.");

    $complete = false;
    for ( $ln=1, $pn=0; ($line=fgets($f)) !== false; $ln++ )
      if ( strlen($line=trim($line)) )
        if ( $line === '[End]' )
        {
          $complete = true;
          break;
        }
        elseif ( ereg( "^([0-9]+)\t(-1|[0-9]+)\t(-?[0-9]+)$", $line, $a ) )
        {
          $prot_index   = $a[1];
          $nearest_prot = $a[2];
          $scoremax     = $a[3];
          if ( $prot_index != $pn )
          {
            add2log( "Error in '$filename' on line $ln: protein number $prot_index ".
                     "doesn't match serial number $pn." );
            return false;
          }
          $pn++;
          $profiles[$the_first][$prot_index][$the_second] =
            array( $nearest_prot, $scoremax );
        }

    fclose($f);

    return $complete;
  }

  // loads protein information headers from .hdr file
  function &load_headers( $filename )
  {
    $headers = array();

    if ( !($f=fopen( $filename, "rb" )) )
      die("Can't open headers file '$filename' for reading.");

    for ( $ln=1, $pn=0; ($line=fgets($f)) !== false; $ln++ )
      if ( strlen($line=trim($line)) )
      {
        if ( ereg( "^(.+) \[[^]]+\]$", $line, $a ) ) // ")) // colorer glitch
          $line = $a[1];
        $headers[$pn++] = $line;
      }

    fclose($f);

    return $headers;
  }

  // calculates minimum of plus-list positions of the given profile
  function minimax( $profile, $plus_list )
  {
    $minimax = false;
    foreach ( $plus_list as $prot_index )
      if ( $minimax === false or $profile[$prot_index][1] < $minimax )
        $minimax = $profile[$prot_index][1];
    return $minimax;
  }

  // calculates maximum of minus-list positions of the given profile
  function maximax( $profile, $minus_list )
  {
    $maximax = false;
    foreach ( $minus_list as $prot_index )
      if ( $maximax === false or $profile[$prot_index][1] > $maximax )
        $maximax = $profile[$prot_index][1];
    return $maximax;
  }

  // calculates given profile quality relative to given plus- and minus-lists
  function profile_quality( $profile, $plus_list, $minus_list )
  {
    $sum_plusminus = 0;
    foreach ( $plus_list as $index )
      $sum_plusminus += $profile[$index][1];
    foreach ( $minus_list as $index )
      $sum_plusminus -= $profile[$index][1];
    $sum_sq = 0;
    foreach ( $plus_list as $index )
      $sum_sq += $profile[$index][1]*$profile[$index][1];
    foreach ( $minus_list as $index )
      $sum_sq += $profile[$index][1]*$profile[$index][1];
    return $sum_plusminus/sqrt($sum_sq*(count($plus_list)+count($minus_list)));
  }

?>
